import { EventEmitter } from 'ee-ts';
import * as JsSip from 'jssip';
import { Log } from '../../log';
import { Session } from './session';
export class SipPhone extends EventEmitter {
    constructor(instanceId, debug) {
        super();
        this.instanceId = instanceId;
        this.type = 'webrtc';
        this.sessionCache = new Map();
        if (debug) {
            JsSip.debug.enable('*');
        }
        else {
            JsSip.debug.disable();
        }
        this.log = new Log();
        this.on('unregistered', () => {
            this.sessionCache.clear();
        });
    }
    async callOption(req = {}) {
        const mediaStream = await this.getUserCallStream(req);
        return {
            sessionTimersExpires: 120,
            pcConfig: req.disableStun
                ? undefined
                : {
                    iceServers: [{ urls: ['stun:stun.l.google.com:19302'] }],
                },
            mediaStream,
        };
    }
    async answer(id, req) {
        if (this.sessionCache.has(id)) {
            const session = this.sessionCache.get(id);
            const params = await this.callOption(req);
            return session.answer(params);
        }
        return false;
    }
    async call(req) {
        const params = {};
        let display = null;
        if (req.params) {
            params.audio = req.params.audio;
            params.video = req.params.video || false;
            params.screen = req.params.screen || false;
            params.disableStun = req.params.disableStun || false;
        }
        if (typeof req.destination !== 'string') {
            throw new Error('bad destination type');
        }
        const invite = await this.callOption(params);
        if (req.params && req.params.display) {
            display = {
                extraHeaders: [`X-Webitel-Display: ${req.params.display}`],
            };
        }
        await this.ua.call(req.destination, {
            ...invite,
            ...display,
        });
    }
    async register(sipConf) {
        const socket = new JsSip.WebSocketInterface(sipConf.server);
        const configuration = {
            realm: sipConf.realm,
            uri: sipConf.uri,
            authorization_user: sipConf.authorization_user,
            ha1: sipConf.ha1,
            user_agent: SipPhone.userAgent,
            sockets: [socket],
            session_timers: true,
            // use_preloaded_route: true,
            display_name: sipConf.display_name,
            register_expires: sipConf.register_expires || 90,
            connection_recovery_min_interval: 5,
            connection_recovery_max_interval: 60,
        };
        const ua = (this.ua = new JsSip.UA(configuration));
        // ua.on('connected', (e: object) => {})
        ua.on('newRTCSession', (e) => {
            const session = e.session;
            const id = session.id;
            const callSession = new Session(e);
            this.storeSession(id, callSession);
            session.on('connecting', () => {
                this.emit('localStreams', callSession, session.connection.getLocalStreams());
            });
            if (session.connection) {
                session.connection.addEventListener('addstream', (evt) => {
                    this.emit('peerStreams', callSession, new Array(evt.stream));
                });
            }
            session.on('peerconnection', (peer) => {
                peer.peerconnection.addEventListener('addstream', async (evt) => {
                    // set remote audio stream
                    this.emit('peerStreams', callSession, new Array(evt.stream));
                });
            });
            session.on('ended', () => {
                // this handler will be called for incoming calls too
                this.removeSession(id, session.connection);
            });
            session.on('failed', () => {
                // this handler will be called for incoming calls too
                this.removeSession(id, session.connection);
            });
            session.on('accepted', () => {
                // the call has answered
                if (!callSession.incoming) {
                    this.emit('peerStreams', callSession, session.connection.getRemoteStreams());
                }
            });
            session.on('confirmed', () => {
                // this handler will be called for incoming calls too
            });
            this.emit('newSession', callSession);
        });
        ua.on('disconnected', (e) => {
            this.emit('unregistered');
        });
        // ua.on('registered', (e: object) => {})
        ua.on('unregistered', (e) => {
            this.emit('unregistered');
        });
        ua.on('registrationFailed', (e) => {
            this.log.error('registrationFailed', e);
        });
        ua.on('registered', (e) => {
            this.emit('registered');
        });
        ua.start();
    }
    async unregister() {
        if (this.ua) {
            this.ua.removeAllListeners();
            await this.ua.stop();
            delete this.ua;
        }
    }
    sipSessionByCallId(id) {
        for (const sess of this.allSession) {
            if (sess.callId === id) {
                return sess;
            }
        }
        return null;
    }
    sipSessionBySipId(id) {
        for (const sess of this.allSession) {
            if (sess.id.startsWith(id)) {
                return sess;
            }
        }
        return null;
    }
    removeSession(id, connection) {
        if (connection) {
            const localStreams = connection.getLocalStreams();
            const remoteStreams = connection.getRemoteStreams();
            if (localStreams) {
                localStreams.forEach((stream) => {
                    stream.getTracks().forEach((track) => {
                        track.stop();
                    });
                });
            }
            if (remoteStreams) {
                remoteStreams.forEach((stream) => {
                    stream.getTracks().forEach((track) => {
                        track.stop();
                    });
                });
            }
        }
        if (this.sessionCache.has(id)) {
            this.sessionCache.delete(id);
            return true;
        }
        return false;
    }
    storeSession(id, session) {
        if (this.sessionCache.has(id)) {
            throw new Error('Session already store');
        }
        this.sessionCache.set(id, session);
    }
    get allSession() {
        return Array.from(this.sessionCache.values());
    }
    getMediaConstraints(req) {
        if (req.screen) {
            return {
                video: false,
                audio: false,
                screen: true,
            };
        }
        return {
            video: req.video || false,
            audio: true,
        };
    }
    getSession(id) {
        if (this.sessionCache.has(id)) {
            return this.sessionCache.get(id);
        }
        return null;
    }
    hasSession(id) {
        return this.sessionCache.has(id);
    }
    async getUserScreen() {
        return new Promise(async (resolve, reject) => {
            try {
                const stream = (await navigator.mediaDevices.getDisplayMedia());
                resolve(stream);
            }
            catch (e) {
                reject(e);
            }
        });
    }
    async getUserCallStream(req) {
        let stream;
        if (req.screen === true) {
            stream = await this.getUserScreen();
        }
        if (req.audio !== false || req.video) {
            const camera = await this.getUserMedia(req.audio, req.video);
            if (stream) {
                camera.getTracks().forEach((track) => {
                    stream.addTrack(track);
                });
            }
            else {
                stream = camera;
            }
        }
        if (!stream) {
            throw new Error('bad stream request');
        }
        return stream;
    }
    async getUserMedia(audio, video) {
        return new Promise(async (resolve, reject) => {
            const mediaConstraints = {
                audio: !(audio === false),
                video: undefined,
            };
            if (video) {
                mediaConstraints.video = {
                    width: { min: 10, ideal: 1280, max: 1920 },
                    height: { min: 10, ideal: 720, max: 1080 },
                    frameRate: { ideal: 10, max: 15 },
                };
            }
            return getMediaStream(mediaConstraints)
                .then(resolve)
                .catch(reject);
        });
    }
}
SipPhone.userAgent = 'Webitel-Phone/0.0.1';
async function getMediaStream(constraints) {
    return new Promise((resolve, reject) => {
        if (navigator.mediaDevices && navigator.mediaDevices.getUserMedia) {
            navigator.mediaDevices
                .getUserMedia(constraints)
                .then((stream) => resolve(stream))
                .catch((err) => reject(err));
        }
        else {
            /// <reference types="webrtc" />
            const getUserMedia = navigator.getUserMedia ||
                navigator.webkitGetUserMedia ||
                navigator.mozGetUserMedia ||
                navigator.msGetUserMedia;
            getUserMedia(constraints, (stream) => resolve(stream), (err) => reject(err));
        }
    });
}
//# sourceMappingURL=index.js.map