// tslint:disable-next-line: match-default-export-name
import Axios from 'axios';
import { EventEmitter } from 'ee-ts';
import { Log } from '../log';
import { SipPhone } from '../sip/webrtc';
import { Agent } from './agent';
import { Call, CallActions, } from './call';
import { ChatActions, Conversation, } from './conversation';
import { Socket } from './socket';
import { ChannelName, Task } from './task';
import { formatBaseUri } from './utils';
const API_HEADER_TOKEN = 'X-Webitel-Access';
const WEBSOCKET_AUTHENTICATION_CHALLENGE = 'authentication_challenge';
const WEBSOCKET_DEFAULT_DEVICE_CONFIG = 'user_default_device';
const WEBSOCKET_AGENT_SESSION = 'cc_agent_session';
const WEBSOCKET_PING = 'ping';
// const WEBSOCKET_CALL_BY_USER = 'call_by_user'
const WEBSOCKET_MAKE_OUTBOUND_CALL = 'call_invite';
const WEBSOCKET_MAKE_USER_CALL = 'call_user';
const WEBSOCKET_EVENT_HELLO = 'hello';
const WEBSOCKET_EVENT_CALL = 'call';
const WEBSOCKET_EVENT_CHAT = 'chat';
const WEBSOCKET_EVENT_USER_STATE = 'user_state';
const WEBSOCKET_EVENT_AGENT_STATUS = 'agent_status';
const WEBSOCKET_EVENT_CHANNEL_STATUS = 'channel';
const WEBSOCKET_EVENT_QUEUE_JOIN_MEMBER = 'queue_join_member';
const TASK_EVENT = 'task';
const WEBSOCKET_EVENT_SIP = 'sip';
export var Response;
(function (Response) {
    Response["STATUS_FAIL"] = "FAIL";
    Response["STATUS_OK"] = "OK";
})(Response || (Response = {}));
export class Client extends EventEmitter {
    constructor(_config) {
        super();
        this._config = _config;
        this.reqSeq = 0;
        this.queueRequest = new Map();
        this.log = new Log();
        this.eventHandler = new EventEmitter();
        this.callStore = new Map();
        this.conversationStore = new Map();
        this.pingTimer = null;
        this.basePath = `${formatBaseUri(_config.storageEndpoint || _config.endpoint)}`;
    }
    async connect() {
        await this.connectToSocket();
    }
    async disconnect() {
        this.socket.close();
    }
    async subscribeCall(handler, data) {
        const calls = (await this.request(`subscribe_call`, data));
        this.eventHandler.on(WEBSOCKET_EVENT_CALL, handler);
        if (calls.items && calls.items.length) {
            for (const c of calls.items) {
                if (c.hangup_at && c.hangup_at > 0 && !c.task) {
                    continue;
                }
                const e = {
                    cc_app_id: '',
                    app_id: c.app_id,
                    event: c.state,
                    id: c.id,
                    timestamp: Date.now(),
                    data: {
                        app_id: c.app_id,
                        event: c.state,
                        id: c.id,
                        timestamp: Date.now(),
                        cc_app_id: '',
                        sip_id: '',
                        parent_id: c.parent_id,
                        direction: c.direction,
                        destination: c.destination,
                        queue: c.queue,
                        from: c.from,
                        to: c.to,
                        payload: c.variables,
                    },
                };
                const call = new Call(this, e);
                call.createdAt = c.created_at;
                call.answeredAt = c.answered_at || 0;
                call.bridgedAt = c.bridged_at || 0;
                call.hangupAt = c.hangup_at || 0;
                call.reportingAt = c.reporting_at || 0;
                if (c.task) {
                    call.task = new Task(this, {
                        attempt_id: c.task.attempt_id,
                        channel: c.task.channel,
                        status: c.task.status,
                        timestamp: Date.now(),
                    }, c.task // todo add app_id
                    );
                    call.queue.reporting = 'true';
                }
                this.callStore.set(call.id, call);
            }
        }
        return calls;
    }
    async subscribeChat(handler, data) {
        const res = (await this.request(`subscribe_chat`, data));
        this.eventHandler.on(WEBSOCKET_EVENT_CHAT, handler);
        if (res && res.items) {
            for (const conv of res.items) {
                // tslint:disable-next-line: no-assign-mutated-array
                const messages = conv.messages.reverse();
                const c = new Conversation(this, conv.id, conv.title, conv.members, messages, conv.variables);
                c.createdAt = conv.created_at;
                if (conv.invite_id) {
                    c.setInvite(conv.invite_id, conv.created_at);
                }
                else if (conv.channel_id) {
                    c.setAnswered(conv.channel_id, conv.joined_at, {
                        id: conv.channel_id,
                        name: 'TODO',
                        type: 'webitel',
                    });
                }
                else {
                    this.log.error(`conversation ${conv.id} not valid`);
                    continue;
                }
                this.conversationStore.set(c.id, c);
            }
        }
        return res;
    }
    async subscribeUsersStatus(handler, data) {
        const res = await this.request(`subscribe_users_status`, data);
        this.eventHandler.on(WEBSOCKET_EVENT_USER_STATE, handler);
        return res;
    }
    async subscribeAgentsStatus(handler, data) {
        const res = await this.request(`cc_agent_subscribe_status`, data);
        this.eventHandler.on(WEBSOCKET_EVENT_AGENT_STATUS, handler);
        return res;
    }
    subscribeTask(handler) {
        this.eventHandler.on(TASK_EVENT, handler);
    }
    async unSubscribe(action, handler, data) {
        const res = await this.request(`un_subscribe_${action}`, data);
        // this.eventHandler.listeners(action)
        // this.eventHandler.removeListener(action, handler)
        // this.eventHandler.off(action, handler)
        return res;
    }
    async destroy() {
        await this.disconnect();
        if (this.phone) {
            await this.phone.unregister();
            this.phone.off('*');
            delete this.phone;
        }
        this.eventHandler.off('*');
    }
    allCall() {
        return Array.from(this.callStore.values());
    }
    allConversations() {
        return Array.from(this.conversationStore.values());
    }
    allTask() {
        if (!this.agent) {
            return [];
        }
        return Array.from(this.agent.task.values());
    }
    callById(id) {
        if (this.callStore.has(id)) {
            return this.callStore.get(id);
        }
    }
    conversationById(id) {
        if (this.conversationStore.has(id)) {
            return this.conversationStore.get(id);
        }
    }
    callBySipId(id) {
        for (const call of this.allCall()) {
            if (call.sipId && id.startsWith(call.sipId)) {
                // FIXME
                return call;
            }
        }
    }
    async reportingTask(attemptId, reporting) {
        return this.request('cc_reporting', {
            attempt_id: attemptId,
            ...reporting,
        });
    }
    async agentActiveTasks(agentId) {
        return this.request(`cc_agent_tasks`, {
            agent_id: agentId,
        });
    }
    async auth() {
        return this.request(WEBSOCKET_AUTHENTICATION_CHALLENGE, {
            token: this._config.token,
        });
    }
    async ping() {
        return this.request(WEBSOCKET_PING, {
            ping: 1,
        });
    }
    sessionInfo() {
        return this.connectionInfo.session;
    }
    get version() {
        return this.connectionInfo.server_version;
    }
    get instanceId() {
        return this.connectionInfo.sock_id;
    }
    async agentSession() {
        const info = await this.request(WEBSOCKET_AGENT_SESSION);
        this.agent = new Agent(this, info);
        const taskList = (await this.agentActiveTasks(this.agent.agentId));
        // todo
        for (const t of taskList.items) {
            const task = new Task(this, {
                status: t.status,
                timestamp: Date.now() - t.duration * 1000,
                channel: t.channel,
                attempt_id: t.attempt_id,
            }, t);
            task.postProcessData = {};
            task.state = t.state;
            if (t.bridged_at) {
                task.bridgedAt = t.bridged_at;
                task.answeredAt = t.bridged_at;
            }
            if (t.leaving_at) {
                task.closedAt = t.leaving_at;
            }
            this.agent.task.set(task.id, task);
        }
        for (const call of this.allCall()) {
            if (call.task) {
                this.agent.task.set(call.task.id, call.task);
            }
        }
        return this.agent;
    }
    async invite(req) {
        return this.request(WEBSOCKET_MAKE_OUTBOUND_CALL, req);
    }
    async call(req) {
        if (this.phone) {
            await this.phone.call(req);
        }
        else {
            await this.invite(req);
        }
    }
    async callRecordId(id) {
        return this.request('call_recordings', { id });
    }
    async eavesdrop(req) {
        return this.request('call_eavesdrop', req);
    }
    async agentSetOnline(agentId, channels, onDemand) {
        return this.request('cc_agent_online', {
            agent_id: agentId,
            channels,
            on_demand: onDemand,
        });
    }
    async agentSetWaiting(agentId, channel) {
        return this.request('cc_agent_waiting', {
            agent_id: agentId,
            channel,
        });
    }
    async agentSetPause(agentId, payload) {
        return this.request('cc_agent_pause', {
            agent_id: agentId,
            payload,
        });
    }
    async agentSetOffline(agentId) {
        return this.request('cc_agent_offline', { agent_id: agentId });
    }
    inviteToUser(req) {
        return this.request(WEBSOCKET_MAKE_USER_CALL, req);
    }
    async answer(id, req) {
        if (this.phone) {
            return this.phone.answer(id, req);
        }
    }
    request(action, data) {
        return new Promise((resolve, reject) => {
            this.queueRequest.set(++this.reqSeq, { resolve, reject });
            this.socket.send({
                seq: this.reqSeq,
                action,
                data,
            });
        });
    }
    useWebPhone() {
        return this._config.registerWebDevice || false;
    }
    callDestroyed(call) {
        return call.hangupAt > 0 && !this.hasAgentTask(call.task);
    }
    conversationDestroyed(conv) {
        return conv.closedAt > 0 && !this.hasAgentTask(conv.task);
    }
    reportingChannelTask(task) {
        switch (task.channel) {
            case ChannelName.Call:
                for (const call of this.allCall()) {
                    if (this.callDestroyed(call)) {
                        this.destroyCall(call);
                        return;
                    }
                }
                break;
            case ChannelName.Chat:
                if (task.agentChannelId) {
                    const conv = this.conversationById(task.agentChannelId);
                    if (conv && this.conversationDestroyed(conv)) {
                        this.destroyConversation(conv);
                        return;
                    }
                }
                break;
            default:
        }
    }
    async registerCallClient(phone) {
        this.phone = phone;
        this.phone.on('peerStreams', (session, streams) => {
            const call = this.callBySession(session);
            if (call && !call.peerStreams.length) {
                call.setPeerStreams(streams);
                this.eventHandler.emit(WEBSOCKET_EVENT_CALL, CallActions.PeerStream, call);
            }
        });
        this.phone.on('localStreams', (session, streams) => {
            const call = this.callBySession(session);
            if (call && !call.localStreams.length) {
                call.setLocalStreams(streams);
                this.eventHandler.emit(WEBSOCKET_EVENT_CALL, CallActions.LocalStream, call);
            }
        });
        this.phone.on('newSession', this.onNewCallSession.bind(this));
        try {
            const conf = await this.deviceConfig(this.phone.type);
            await this.phone.register(conf);
        }
        catch (e) {
            // FIXME add handle error
            this.log.error(e);
        }
    }
    async deviceConfig(name) {
        return this.request(WEBSOCKET_DEFAULT_DEVICE_CONFIG, { name });
    }
    async storeFile(id, files, cb) {
        if (!files || files.length < 1) {
            throw new Error('no files');
        }
        const formData = new FormData();
        for (const file of files) {
            formData.append(file.name, file); // todo name
        }
        const result = await Axios.post(`${this.basePath}/api/storage/file/${id}/upload`, formData, {
            headers: {
                [API_HEADER_TOKEN]: this._config.token,
            },
            onUploadProgress: cb,
        });
        const data = result.data;
        for (const f of data) {
            f.shared = this.basePath + f.shared;
        }
        return data;
    }
    fileUrlDownload(fileId) {
        return `${this.basePath}/api/storage/file/${fileId}/download?access_token=${this._config.token}`;
    }
    fileUrlStream(fileId) {
        return `${this.basePath}/api/storage/file/${fileId}/stream?access_token=${this._config.token}`;
    }
    async onMessage(message) {
        this.log.debug('receive message: ', message);
        if (message.seq_reply > 0) {
            if (this.queueRequest.has(message.seq_reply)) {
                const promise = this.queueRequest.get(message.seq_reply);
                this.queueRequest.delete(message.seq_reply);
                if (message.status === Response.STATUS_OK) {
                    promise.resolve(message.data);
                }
                else {
                    promise.reject(message.error);
                }
            }
        }
        else {
            switch (message.event) {
                case WEBSOCKET_EVENT_HELLO:
                    await this.connected(message.data);
                    this.log.debug(`opened session ${this.connectionInfo.sock_id} for userId=${this.connectionInfo.session.user_id}`);
                    break;
                case WEBSOCKET_EVENT_CALL:
                    await this.handleCallEvents(message.data.call);
                    break;
                case WEBSOCKET_EVENT_CHAT:
                    await this.handleChatEvents(message.data);
                    break;
                case WEBSOCKET_EVENT_USER_STATE:
                    this.handleUserStateEvent(message.data.state);
                    break;
                case WEBSOCKET_EVENT_SIP:
                    this.eventHandler.emit(WEBSOCKET_EVENT_SIP, message.data);
                    break;
                case WEBSOCKET_EVENT_AGENT_STATUS:
                    this.handleAgentStatus(message.data);
                    break;
                case WEBSOCKET_EVENT_CHANNEL_STATUS:
                    this.handleChannelEvents(message.data);
                    break;
                default:
                    this.log.error(`event ${message.event} not handler`);
            }
        }
    }
    handleAgentStatus(e) {
        if (this.agent && e.agent_id === this.agent.agentId) {
            this.agent.setStatus(e);
        }
        this.eventHandler.emit(WEBSOCKET_EVENT_AGENT_STATUS, e, this.agent);
    }
    handleChannelEvents(e) {
        if (this.agent) {
            e.timestamp = Date.now(); // bug
            const task = this.agent.onChannelEvent(e) || undefined;
            this.eventHandler.emit(TASK_EVENT, e.status, task);
        }
    }
    async pingServer() {
        if (this.pingTimer) {
            await this.ping();
        }
        if (this.connectionInfo.ping_interval) {
            // @ts-ignore
            this.pingTimer = setTimeout(this.pingServer.bind(this), +this.connectionInfo.ping_interval);
        }
    }
    async connected(info) {
        this.connectionInfo = info;
        if (info.ping_interval) {
            await this.pingServer();
        }
        if (!this.useWebPhone()) {
            return;
        }
        return this.registerCallClient(new SipPhone(this.instanceId, this._config.debug));
    }
    callBySession(session) {
        for (const call of this.allCall()) {
            if (call.sip === session) {
                return call;
            }
        }
    }
    async onNewCallSession(session) {
        let call;
        if (session.callId) {
            call = this.callById(session.callId);
        }
        else {
            call = this.callBySipId(session.id);
        }
        if (call) {
            call.setSip(session);
            this.checkAutoAnswer(call);
        }
    }
    checkAutoAnswer(call) {
        if ( /*!document.hidden && */call.autoAnswer) {
            call.answerDelay({
                video: call.params.video,
                screen: call.params.screen,
                disableStun: call.params.disableStun,
            }, 300);
        }
    }
    connectToSocket() {
        return new Promise((resolve, reject) => {
            try {
                this.socket = new Socket(this._config.endpoint);
                this.socket.connect(this._config.token);
            }
            catch (e) {
                reject(e);
                return;
            }
            this.socket.on('message', this.onMessage.bind(this));
            this.socket.on('close', (code) => {
                if (this.pingTimer) {
                    clearTimeout(this.pingTimer);
                }
                this.log.error('socket close code: ', code);
                this.eventHandler.off('*');
                if (code !== 1000) {
                    this.emit('disconnected', code);
                    reject(new Error(`close socket code: ${code}`));
                }
            });
            this.socket.on('open', () => {
                resolve(null);
                this.emit('connected');
            });
        });
    }
    handleUserStateEvent(event) {
        this.eventHandler.emit(WEBSOCKET_EVENT_USER_STATE, event);
    }
    async handleCallEvents(event) {
        let call;
        event.timestamp = Date.now(); // todo bug
        switch (event.event) {
            case CallActions.Ringing:
                call = new Call(this, event);
                this.callStore.set(call.id, call);
                this.checkAutoAnswer(call);
                break;
            case CallActions.Active:
                call = this.callById(event.id);
                if (call) {
                    call.setActive(event);
                }
                break;
            case CallActions.Bridge:
                call = this.callById(event.id);
                if (call) {
                    call.setBridged(event);
                }
                break;
            case CallActions.Execute:
                call = this.callById(event.id);
                if (call) {
                    call.setExecute(event.data);
                }
                break;
            case CallActions.DTMF:
                call = this.callById(event.id);
                if (call) {
                    call.addDigit(event.data);
                }
                break;
            case CallActions.Voice:
                call = this.callById(event.id);
                if (call) {
                    call.setVoice();
                }
                break;
            case CallActions.Silence:
                call = this.callById(event.id);
                if (call) {
                    call.setSilence();
                }
                break;
            case CallActions.Hold:
                call = this.callById(event.id);
                if (call) {
                    call.setHold(event);
                }
                break;
            case CallActions.Hangup:
                call = this.callById(event.id);
                if (call) {
                    call.setHangup(event);
                }
                break;
            default:
                throw new Error('Unhandled action');
        }
        if (call) {
            this.eventHandler.emit(WEBSOCKET_EVENT_CALL, event.event, call);
            if (this.callDestroyed(call)) {
                this.destroyCall(call);
            }
        }
    }
    async handleChatEvents(event) {
        let conversation;
        const timestamp = Date.now(); // todo bug
        switch (event.action) {
            case ChatActions.UserInvite:
                const inv = event.data;
                conversation = new Conversation(this, inv.conversation_id, inv.title, inv.members, inv.messages, inv.variables);
                conversation.setInvite(inv.invite_id, timestamp);
                this.conversationStore.set(conversation.id, conversation);
                break;
            case ChatActions.Joined:
                const joined = event.data;
                conversation = this.conversationById(joined.conversation_id);
                if (conversation) {
                    conversation.setAnswered(joined.member.id, timestamp, joined.member);
                }
                break;
            case ChatActions.Message:
                const message = event.data;
                message.timestamp = timestamp;
                conversation = this.conversationById(message.conversation_id);
                if (conversation) {
                    conversation.newMessage(message);
                }
                break;
            case ChatActions.Close:
            case ChatActions.Leave:
            case ChatActions.Decline:
                // TODO leave & CLOSE:
                const e = event.data;
                conversation = this.conversationById(e.conversation_id);
                if (conversation) {
                    conversation.setClosed(timestamp);
                }
                break;
            default:
        }
        if (conversation) {
            this.eventHandler.emit(WEBSOCKET_EVENT_CHAT, event.action, conversation);
            if (this.conversationDestroyed(conversation)) {
                this.destroyConversation(conversation);
            }
        }
    }
    hasAgentTask(task) {
        if (!this.agent || !task) {
            return false;
        }
        return this.agent.hasTask(task);
    }
    destroyCall(call) {
        // FIXME sync channel & call event
        this.callStore.delete(call.id);
        this.eventHandler.emit(WEBSOCKET_EVENT_CALL, CallActions.Destroy, call);
    }
    destroyConversation(conv) {
        // FIXME sync channel & call event
        this.conversationStore.delete(conv.id);
        this.eventHandler.emit(WEBSOCKET_EVENT_CHAT, ChatActions.Destroy, conv);
    }
}
//# sourceMappingURL=client.js.map